// ASNFileParse.cpp: implementation of the CASNFileParse class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ASN1Decode.h"
#include "ASNFileParse.h"
#include "tbase64.h"
#include "ASN.h"
#include "obj_dat.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

static char* GetTagName(char cTag)
{
	switch(cTag)
	{
	case TAG_BOOL:			return "BOOLEAN";
	case TAG_INTEGER:		return "INTEGER";
	case TAG_STR_BIT:		return "BIT STRING";					// λַ
	case TAG_STR_OCT:		return "OCTET STRING";
	case TAG_NULL:			return "NULL";
	case TAG_OIDOBJ:		return "OBJECT";
	//18-220x12-0x16, 25-300x19-0x1e ַ
	case TAG_STR_UTF8:		return "UTF8 STRING";
	case TAG_STR_PRINTABLE:	return "PRINTABLE STRING";

	case TAG_UTC_TIME:		return "UTC TIME";
	case TAG_SEQUENCE:		return "SEQUENCE";
	case TAG_SET:			return "SET";

	case TAG_EXTENTION_A0:	return "EXTENSION";		// ֤汾
	case TAG_EXTENTION_A1:	return "EXTENSION";		// ֤鷢id
	case TAG_EXTENTION_A2:	return "EXTENSION";		// ֤id
	case TAG_EXTENTION_A3:	return "EXTENSION";		// ֤չֶ
	default:
		return "";
	}
}
BOOL CASNFileParse::LogOut(char* pBuff, int iBuffSize, int& iBuffPos, const char* pszFormat, ...)
{
	if(NULL == pBuff)
	{
		return FALSE;
	}

	char szBuff [4096] = {0};
	va_list pArgList ;
	va_start (pArgList, pszFormat) ;
	(void)_vsnprintf (szBuff, sizeof (szBuff) / sizeof (char)-1, 
		pszFormat, pArgList) ;
	va_end (pArgList);

	int iLen = strlen(szBuff);
	if(iBuffSize-iBuffPos < iLen)
		return FALSE;

	memcpy(pBuff+iBuffPos, szBuff, iLen);
	iBuffPos += iLen;
	return TRUE;
}


// ASNļָļ
void CASNFileParse::OutputASN2Buff(char* pBuff, int iBuffSize, int& iBuffPos, CAsnSetX* pNodeSetX, int iLevel)
{
	ASSERT(pNodeSetX);
	if( NULL == pNodeSetX)
	{
		return;
	}

	const CString cstrSeparator = "\t";
	CString cstrSpace;
	for(int i = 0; i < iLevel; i++)
	{
		cstrSpace += cstrSeparator;	
	}

	const int INT_LINENUM = 32;		// ÿʾַ
	BYTE cTag = pNodeSetX->GetTag();
	DWORD dwDataSize = 0;
	DWORD dwTagSize = 0;
	pNodeSetX->GetSize(dwTagSize, dwDataSize);

	LogOut(pBuff, iBuffSize, iBuffPos,"%s{TAG(0x%02x):%s, SIZE: %d\r\n", cstrSpace, cTag, GetTagName(cTag), dwDataSize);
	// ǣɾԭļ

	CTypedPtrList<CPtrList, IAsnNode*>& lstSet = pNodeSetX->GetList();

	for(POSITION pos = lstSet.GetHeadPosition(); pos; )
	{
		IAsnNode* pNode = lstSet.GetNext(pos);
		if(pNode->IsStruct())
		{
			CAsnSetX* pNodeSetX = static_cast<CAsnSetX*>(pNode);
			OutputASN2Buff(pBuff, iBuffSize, iBuffPos, pNodeSetX, iLevel+1);
		}
		else 
		{
			CAsnX* pNodeX = static_cast<CAsnX*>(pNode);
			int cTag = pNodeX->GetTag();
			BYTE* pData = pNodeX->GetData();
			DWORD dwTagSize = pNodeX->GetTagSize();
			DWORD dwDataSize = pNodeX->GetDataSize();
			CString cstrData = DataToHexStr(pData, dwDataSize);

			// дļ
			LogOut(pBuff, iBuffSize, iBuffPos, "\r\n%s%s[TAG(0x%02x):%s, SIZE: %d]\r\n", cstrSeparator, cstrSpace, cTag, GetTagName(cTag), dwDataSize);
			if(TAG_OIDOBJ == cTag)
			{
				CString cstrOID = HexToOIDStr(pData, dwDataSize);
				CString cstrSN = CASNFileParse::GetObjNameByOID(cstrOID);
				LogOut(pBuff, iBuffSize, iBuffPos, "%s%s%s(%s) %s\r\n", cstrSeparator, cstrSpace, cstrData, cstrOID, cstrSN);	
			}
			else if((cTag >= 0x12 && cTag <=0x16) || 0xc == cTag)
			{
				char* pStr = new char[dwDataSize+1];
				memcpy(pStr, pData, dwDataSize);
				pStr[dwDataSize] = 0;
				LogOut(pBuff, iBuffSize, iBuffPos, "%s%s%s(%s)\r\n", cstrSeparator, cstrSpace, cstrData, pStr);	
				delete pStr;
			}
			else 
			{
				LogOut(pBuff, iBuffSize, iBuffPos, "%s%s", cstrSeparator, cstrSpace);

				for(DWORD iIdx = 1; iIdx <= dwDataSize; iIdx++)
				{
					CHAR szHex[5] = {0};  
					sprintf(szHex, "%02x", (BYTE)pData[iIdx-1]); 
					if(iIdx % INT_LINENUM == 0)
					{
						// һַ
						if(iIdx != dwDataSize)
							LogOut(pBuff, iBuffSize, iBuffPos, "%s\r\n%s%s", szHex, cstrSeparator, cstrSpace);	
						else
							LogOut(pBuff, iBuffSize, iBuffPos,  "%s", szHex);		
					}
					else
					{
						LogOut(pBuff, iBuffSize, iBuffPos,   "%s ", szHex);
					}

				}
				LogOut(pBuff, iBuffSize, iBuffPos, "\r\n");
			}
		} // end if
	} // end for

	LogOut(pBuff, iBuffSize, iBuffPos, "%s}%dL\r\n", cstrSpace, iLevel);
}


static const ASN1_OBJECT* GetASNObjByOID(const CString& cstrOID)
{
	DWORD dwOIDSize = 0;
	BYTE* pOID = OIDStrToHex(cstrOID, dwOIDSize);
	if(NULL == pOID)
		return NULL;

	for(int i = 0; i < NUM_NID; i++)
	{
		// жϳ
		if(nid_objs[i].length != dwOIDSize)
			continue;
		const unsigned char* pObjData = nid_objs[i].data;
		if(0 == memcmp(pObjData, pOID, dwOIDSize))
		{
			delete pOID;
			return &nid_objs[i];
		}
	}
	delete pOID;
	return NULL;
}
// OIDַȡƣ"1.2.840.113549.3.7" > "DES-EDE3-CBC"
LPCTSTR CASNFileParse::GetObjNameByOID(const CString& cstrOID, BOOL bSN)
{
	/*int iNID = OBJ_txt2nid(cstrOID);
	return bSN ? OBJ_nid2sn(iNID) : OBJ_nid2ln(iNID);*/
	const ASN1_OBJECT* pObj = GetASNObjByOID(cstrOID);
	if(!pObj)
		return "";
	return bSN ? pObj->sn : pObj->ln;
	
}


/*
	Proc-Type: 4,ENCRYPTED
	DEK-Info: DES-EDE3-CBC,BCC508172ACB6E99
*/
static BOOL PickValue(const CString& cstrValue, CString& cstrValue1, CString& cstrValue2)
{
	int iPos = cstrValue.Find(',');
	if(iPos < 0)
		return FALSE;
	cstrValue1 = cstrValue.Left(iPos);
	cstrValue2 = cstrValue.Mid(iPos+1);
	cstrValue1.TrimRight();
	cstrValue2.TrimLeft();
	return TRUE;
}
// PEMͷ
static BOOL PEMHeaderParse(CStringArray& arrPEMHeader, BOOL& bEncrypted, CString& cstrDES, CString& cstrIV)
{
	for(int i = 0; i < arrPEMHeader.GetSize(); i++)
	{
		CString cstrLine = arrPEMHeader[i];
		if(cstrLine.IsEmpty())
			continue;
		int iIndex = cstrLine.Find(':');
		if(iIndex < 0)
			continue;
	
		CString cstrKey = cstrLine.Left(iIndex);
		CString cstrValue = cstrLine.Mid(iIndex + 1);
		cstrKey.TrimRight();
		cstrValue.TrimLeft();
		if(0 == cstrKey.CompareNoCase(PH_PROC_TYPE))
		{
			CString cstrValue1, cstrValue2;
			if(!PickValue(cstrValue, cstrValue1, cstrValue2))
				continue;
			bEncrypted = (0 == cstrValue2.CompareNoCase(PH_ENCRYPTED));
		}
		else if(0 == cstrKey.CompareNoCase(PH_DEK_INFO))
		{
			if(!PickValue(cstrValue, cstrDES, cstrIV))
				continue;
		}
	}
	return TRUE;
}

// PEMʽ
BOOL CASNFileParse::PEM2AsnBinary(const char* pszData, BYTE** pOutData, DWORD& dwOutDataLen, LPCTSTR pszPassword)
{
	*pOutData = NULL;
	dwOutDataLen = 0;

	if(!pszData)
		return FALSE;

	int iLen = strlen(pszData);
	CString cstrFullLine = pszData;
	cstrFullLine.Replace("\r", "");
	// ȷһܱ
	if(cstrFullLine.GetAt(iLen-1) != '\n')
		cstrFullLine += "\n";

	CStringArray arrBody;
	int iStatus = STATUS_PEM_NONE;
	int iHeadLineIdx = 0;

	BOOL bPKCS8 = FALSE;	// PEMļǷ
	// нз
	char szSep[] = "\n";			
	int iSepLen = strlen(szSep);
	while(!cstrFullLine.IsEmpty())
	{
		int iIndex = cstrFullLine.Find(szSep);		// 
		if (iIndex < 0)
			break;

		// ȡһнз
		CString cstrLine = cstrFullLine.Left(iIndex);
		cstrLine.TrimLeft();
		cstrLine.TrimRight();
		
		cstrFullLine = cstrFullLine.Mid(iIndex + iSepLen);
		// жϵǰǷΪ
		if(cstrLine.IsEmpty() && STATUS_PEM_NONE == iStatus)
		{
			continue;
		}
		// жǷʼ
		if(0 == cstrLine.Find(STR_PEM_BEGIN))
		{
			if(0 == cstrLine.CompareNoCase(STR_PRIVKEY_PKCS8_BEGIN))
				bPKCS8 = TRUE;
			else if(0 == cstrLine.CompareNoCase(STR_PRIVKEY_PKCS8_EN_BEGIN))
				bPKCS8 = TRUE;

			iStatus = STATUS_PEM_BEGIN;
		}
		else if(0 == cstrLine.Find(STR_PEM_END))
		{
			iStatus = STATUS_PEM_END;
			break;
		}
		else
		{
			if(cstrLine.IsEmpty())
				iHeadLineIdx = arrBody.GetSize();
			arrBody.Add(cstrLine);
		}
	}

	// PEMͷPEMݷ
	int i = 0;
	CStringArray arrPEMHeader, arrPEMData;
	for(i = 0; i < iHeadLineIdx; i++)
	{
		arrPEMHeader.Add(arrBody[i]);
		TRACE("%d : %s\r\n", i, arrBody[i]);
	}
	for(i = iHeadLineIdx; i < arrBody.GetSize(); i++)
	{
		arrPEMData.Add(arrBody[i]);
	}

	// ڴ
	BYTE* pOut = new BYTE[iLen];
	DWORD dwOutLen = 0;
	for(i = 0; i < arrPEMData.GetSize(); i++)
	{
		CString cstrLine = arrPEMData[i];
		cstrLine.TrimLeft();
		cstrLine.TrimRight();
		if(cstrLine.IsEmpty())
			continue;

		//TRACE("%s\r\n", cstrLine);
		int iDecodeLen = 0;
		// base64
		BYTE* pDecode = base64_decode(cstrLine, strlen(cstrLine), &iDecodeLen);
		if(NULL == pDecode)
		{
			delete pOut; 
			return FALSE;
		}
		memcpy(pOut+dwOutLen, pDecode, iDecodeLen);
		dwOutLen += (DWORD)iDecodeLen;
		free(pDecode);
	}
	*pOutData = pOut;
	dwOutDataLen = dwOutLen;
	
	// ==================ж==================
	// ûм
	if(arrPEMHeader.GetSize() <= 0 && !bPKCS8)
		return TRUE;
	
	// PEMͷ鿴ϢǷ㷨IV
	BOOL bEncrypted = FALSE;
	CString cstrCipherName, cstrIV;
	PEMHeaderParse(arrPEMHeader, bEncrypted, cstrCipherName, cstrIV);
	if(!bEncrypted)
		return TRUE;

	m_bEncrypt = TRUE;
	// DES3ܣܺݲASN.1ʽ
	//CString cstrPassword = pszPassword ? pszPassword : "";	
	//void *pCipherFunc = GetCipherFuncByName(cstrCipherName);
	//if(!DES_Decrypt(pOut, dwOutLen, pOutData, dwOutDataLen, cstrPassword, cstrIV, pCipherFunc))
	//{
	//	delete pOut;
	//	*pOutData = NULL;
	//	dwOutDataLen = 0;
	//	return FALSE;
	//}
	//
	//delete pOut;
	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CASNFileParse::CASNFileParse()
{
	m_pAsnSet = new CAsnSetX(TAG_SEQUENCE);
	m_bEncrypt = FALSE;
	
}

CASNFileParse::~CASNFileParse()
{
	delete m_pAsnSet;
	m_pAsnSet = NULL;

}

// زļ
BOOL CASNFileParse::LoadFile(LPCTSTR pszPathFile, LPCTSTR pszPassword, EM_CERT_TYPE emType)
{
	m_emFileType = emType;

	// ļչжʲôʽļ
	CString cstrExt;
	CString cstrPathFile = pszPathFile;
	int iPos = cstrPathFile.ReverseFind('.');
	if(iPos > 0)
		cstrExt = cstrPathFile.Mid(iPos+1);
	if(0 == cstrExt.CompareNoCase("p12"))			// p12׺ = PKCS#12
	{
		m_emFileType = EM_KEY_PKCS12;
	}
	else if(0 == cstrExt.CompareNoCase("pfx"))		// pfx׺΢׺Ϣ
	{
		m_emFileType = EM_KEY_PKCS12;
	}

	// ļ
	CFile file;
	if(!file.Open(pszPathFile, CFile::modeRead|CFile::shareDenyNone|CFile::typeBinary))
		return FALSE;

	DWORD dwFileLen = file.GetLength();
	char* pszFileData = new char[dwFileLen+1];
	file.Read(pszFileData, dwFileLen);
	file.Close();	
	
	*(pszFileData+dwFileLen) = 0;
	char szPEM[11] = {0};
	// PEMʽıʾ
	if(dwFileLen > 10)
		memcpy(szPEM, pszFileData, 10);
	else
		return FALSE;

	// жļ
	BOOL bRet = TRUE;
	if(0 == stricmp(STR_PEM_BEGIN, szPEM))
	{
		if(0 < strstr(pszFileData, STR_CERTIFACATE_BEGIN))				// ֤ļ
		{
			m_emFileType = EM_CERTIFACATE;
		}
		else if(0 < strstr(pszFileData, STR_PRIVKEY_RSA_BEGIN))			// ˽Կļ
		{
			m_emFileType = EM_PRIVATEKEY_PKCS1;
		}
		else if(0 < strstr(pszFileData, STR_PRIVKEY_PKCS8_BEGIN))		// ˽ԿļPKCS8
		{
			m_emFileType = EM_PRIVATEKEY_PKCS8;
		}
		else if(0 < strstr(pszFileData, STR_PRIVKEY_PKCS8_EN_BEGIN))	// ˽ԿļPKCS8ܣ
		{
			m_emFileType = EM_PRIVATEKEY_PKCS8_ENCRYPT;
		}
		else if(0 < strstr(pszFileData, STR_PUBKEY_BEGIN))				// Կļ
		{
			m_emFileType = EM_PUBLICKEY;
		}

		if(!(bRet = LoadPEMData(pszFileData, m_emFileType, pszPassword)))
		{
			delete pszFileData;
			return FALSE;
		}
	}
	else
	{
		// ƣPEMʽ
		m_pAsnSet->Clear();
		m_pAsnSet->Decode((BYTE*)pszFileData);

	}

	delete pszFileData;
	return bRet;
}

// ض
BOOL CASNFileParse::LoadBinaryData(BYTE* pData, int iDataLen, EM_CERT_TYPE emType, LPCTSTR pszPassword)
{
	if(NULL == pData)
		return FALSE;

	// ƣPEMʽ
	m_pAsnSet->Clear();
	m_pAsnSet->Decode((BYTE*)pData);

	m_emFileType = emType;
	return TRUE;
}

// PEM
BOOL CASNFileParse::LoadPEMData(LPCTSTR pcszData, EM_CERT_TYPE emType, LPCTSTR pszPassword)
{
	// base64
	BYTE* pOutData = NULL;
	DWORD dwOutLen = 0;
	m_emFileType = emType;
	if(!PEM2AsnBinary(pcszData, &pOutData, dwOutLen, pszPassword))
	{
		delete pOutData;
		return FALSE;
	}
	
	m_pAsnSet->Clear();
	if(!m_pAsnSet->Decode(pOutData))
	{
		delete pOutData;
		return FALSE;
	}
	delete pOutData;
	return TRUE;
}


